#include <QWidget>
#include <QVBoxLayout>
#include <QLineEdit>
#include <QPushButton>
#include <QTextEdit>
#include <QLabel>
#include <QString>
#include <QThread>
#include <QCheckBox>
#include <QHBoxLayout>
#include <cmath>
#include <QSpinBox>
#include <QApplication>

using namespace std;
using namespace std::chrono;

int partition(vector<int>& arr, int low, int high) {
    int pivot = arr[high];
    int i = low - 1;
    for (int j = low; j <= high - 1; j++)
        if (arr[j] < pivot) {
            i++;
            swap(arr[i], arr[j]);
        }
    swap(arr[i + 1], arr[high]);  
    return i + 1;
}
void quickSort(vector<int>& arr, int low, int high) {
    
    if (low < high) {
        int pi = partition(arr, low, high);
        quickSort(arr, low, pi - 1);
        quickSort(arr, pi + 1, high);
    }
}

int partition(std::vector<std::array<char, 27>>& arr, int low, int high) {
    const char* pivot = &arr[high][1];  // Вибираємо опорний елемент, ігноруючи перший символ
    int i = low - 1;
    
    for (int j = low; j <= high - 1; j++) {
        if (strcmp(&arr[j][1], pivot) < 0) {  // Порівнюємо рядки з другого символу
            i++;
            std::swap(arr[i], arr[j]);
        }
    }
    std::swap(arr[i + 1], arr[high]);
    return i + 1;
}

void quickSort(std::vector<std::array<char, 27>>& arr, int low, int high) {
    if (low < high) {
        int pi = partition(arr, low, high);
        quickSort(arr, low, pi - 1);
        quickSort(arr, pi + 1, high);
    }
}
class SortWorker : public QObject {
    Q_OBJECT
public:
    SortWorker(vector<int>& arr, unsigned int n, bool individualTask) 
        : arr(arr), n(n), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        auto start_time = high_resolution_clock::now();
        quickSort(arr,0,arr.size()-1);
        auto duration = duration_cast<microseconds>(high_resolution_clock::now() - start_time);
        emit progressUpdated(100, n, n);
        emit sortingFinished(duration.count(), is_sorted(arr.data(), n, [](int a, int b) { return a < b; }));
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<int>& arr;
    unsigned int n;
    bool individualTask;
    std::atomic<bool> shouldStop;
    template<typename Compare>
    bool is_sorted(int* arr, unsigned int n, Compare comp) {
        for (unsigned int i = 0; i < n - 1; i++) {
            if (comp(arr[i+1], arr[i])) return false;
        }
        return true;
    }
};
class SortWorkerWorse : public QObject {
    Q_OBJECT
public:
    SortWorkerWorse(vector<std::array<char, 27>>& arr, unsigned int n, bool individualTask) 
        : arr(arr), individualTask(individualTask), shouldStop(false) {}
    
    void stop() {
        shouldStop = true;
    }
    
public slots:
    void performSort() {
        auto start_time = high_resolution_clock::now();
        quickSort(arr,0,arr.size()-1);
        auto duration = duration_cast<microseconds>(high_resolution_clock::now() - start_time);
        emit progressUpdated(100, arr.size(), arr.size());
        emit sortingFinished(duration.count(),1);
    }
    
signals:
    void progressUpdated(int progress, unsigned int current, unsigned int total);
    void sortingFinished(int duration, bool sorted);
    void sortingCancelled();
    
private:
    vector<std::array<char, 27>>& arr;
    bool individualTask;
    std::atomic<bool> shouldStop;
};

class SortingApp : public QWidget {
    Q_OBJECT
public:
    SortingApp(QWidget *parent = nullptr) : QWidget(parent), worker(nullptr), wworker(nullptr), isSorting(false) {
        setupUI();
    }

    ~SortingApp() {
        if (workerThread.isRunning()) {
            workerThread.quit();
            workerThread.wait();
        }
    }

private:
    QSpinBox *inputField;
    QPushButton *sortButton;
    QTextEdit *outputField;
    QCheckBox *showArrayCheckBox;
    vector<int> arr;
    vector<std::array<char, 27>> arrI;//mark+last[15]+first[10]+0
    QThread workerThread;
    SortWorker *worker;
    SortWorkerWorse *wworker;
    QCheckBox *individualTaskCheckBox;
    bool isSorting;

    void setupUI() {
        QVBoxLayout *mainLayout = new QVBoxLayout(this);

        QLabel *label = new QLabel("Скільки чисел сортувати?", this);
        mainLayout->addWidget(label);

        inputField = new QSpinBox(this);
        inputField->setRange(1, 10000000);
        inputField->setValue(10);  // Встановлюємо початкове значення
        mainLayout->addWidget(inputField);

        sortButton = new QPushButton("Сортувати", this);
        connect(sortButton, &QPushButton::clicked, this, &SortingApp::onSortButtonClicked);
        mainLayout->addWidget(sortButton);

        QHBoxLayout *checkBoxLayout = new QHBoxLayout();
        
        showArrayCheckBox = new QCheckBox("Показати масив до та після сортування", this);
        checkBoxLayout->addWidget(showArrayCheckBox);

        individualTaskCheckBox = new QCheckBox("Індивідуальне завдання", this);
        checkBoxLayout->addWidget(individualTaskCheckBox);

        mainLayout->addLayout(checkBoxLayout);

        outputField = new QTextEdit(this);
        outputField->setReadOnly(true);
        mainLayout->addWidget(outputField);

        setLayout(mainLayout);
        setWindowTitle("Сортування масиву");
        resize(500, 500);
    }

    void sortArray() {
        arr.clear();
        arrI.clear();
        if (individualTaskCheckBox->isChecked()){
            for (unsigned int i = 0; i < inputField->value(); i++){
                std::array<char, 27> pushme;
                for (char j = 0; j < 26; j++)
                    if (!j)pushme[j]=rand()%5+1;
                    else if (j==1||j==16)pushme[j] = rand()%26+65;else if (j==15)pushme[j]=32;else pushme[j]=rand()%26+97;
                pushme[26]=0;
            arrI.push_back(pushme);}
        }
        else{
        arr.resize(inputField->value());
        for (unsigned int i = 0; i < inputField->value(); i++) arr[i] = rand() % 20000001 - 10000000;
        }
        outputField->clear();
        if (showArrayCheckBox->isChecked()) {
            outputField->append("Масив до сортування:");
            if (individualTaskCheckBox->isChecked()) outputField->append(vvToString(arrI));
            else outputField->append(arrayToString(arr));
        }
        if (individualTaskCheckBox->isChecked()){
            wworker = new SortWorkerWorse(arrI, inputField->value(), individualTaskCheckBox->isChecked());
            wworker->moveToThread(&workerThread);
            outputField->append("==Режим індивідуального завдання #8==\n");
            for (unsigned int b=0;b<arrI.size();b++)
                if (arrI[b][0]<4){arrI.erase(arrI.begin()+b);b--;}
            connect(this, &SortingApp::startSort, wworker, &SortWorkerWorse::performSort);
            connect(wworker, &SortWorkerWorse::progressUpdated, this, &SortingApp::updateProgress);
            connect(wworker, &SortWorkerWorse::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, wworker, &QObject::deleteLater);
        }else{
            worker = new SortWorker(arr, inputField->value(), individualTaskCheckBox->isChecked());
            worker->moveToThread(&workerThread);
            connect(this, &SortingApp::startSort, worker, &SortWorker::performSort);
            connect(worker, &SortWorker::progressUpdated, this, &SortingApp::updateProgress);
            connect(worker, &SortWorker::sortingFinished, this, &SortingApp::finishSorting);
            connect(&workerThread, &QThread::finished, worker, &QObject::deleteLater);}
        workerThread.start();
        emit startSort();
    }
    
    QString arrayToString(const vector<int>& arr) {
        QString result;
        for (size_t i = 0; i < arr.size(); ++i) {
            result += QString::number(static_cast<int>(arr[i]));
            if (i < arr.size() - 1) result += ", ";
        }
        return result;
    }
    QString vvToString(const vector<std::array<char, 27>>& arr) {
        QString result;
        for (size_t i = 0; i < arr.size(); ++i){ for (char a=1;a<26;a++)result+=arr[i][a];
            result+=": "+QString::number(arr[i][0]);
            if (i!=arr.size()-1)result+='\n';
        }
        return result;
    }

private slots:
    void onSortButtonClicked() {
        if (!isSorting) {
        isSorting = true;
        sortButton->setText("Скасувати");
            sortArray();
        } else resetSortingState();
    }

    void updateProgress(int progress, unsigned int current, unsigned int total) {
        outputField->append(QString("Відсортовано %1% (%2 з %3 елементів)").arg(progress).arg(current).arg(total));
    }

    void finishSorting(int duration, bool sorted) {
        outputField->append(QString("\nЧас сортування: %1 мікросекунд").arg(duration));
        outputField->append(sorted ? "Масив відсортовано правильно." : "Помилка: масив не відсортовано правильно.");
        
        if (showArrayCheckBox->isChecked()) {
            outputField->append("\nМасив після сортування:");
            if (individualTaskCheckBox->isChecked())outputField->append(vvToString(arrI));else
            outputField->append(arrayToString(arr));
            arr.clear();
        }
        
        resetSortingState();
    }
    void resetSortingState() {
        isSorting = false;
        sortButton->setText("Сортувати");
        if(worker){
            worker->disconnect();worker->deleteLater();}
        if(wworker){wworker->deleteLater();
            wworker->disconnect();}
        worker=nullptr;
        wworker=nullptr;
    }

signals:
    void startSort();
};

int main(int argc, char *argv[]) {
    QApplication app(argc, argv);
    SortingApp sortingApp;
    sortingApp.show();
    return app.exec();
}
#include "main.moc"
